#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <getopt.h>
#include <limits.h>
#include <dirent.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/wait.h>

#define MUPIP_CMD "mupip"
#define MUMPS_CMD "mumps"

char mupip[PATH_MAX+1];
char mumps[PATH_MAX+1];
int quiet = 0;

#define ERR_PRINT(...) if (!quiet) fprintf( stderr, __VA_ARGS__ )

int is_int( char *str ) {
  long long_val;
  char *end_ptr;
  if (!str) {
    return -1;
  }
  end_ptr = str;
  long_val = strtol( str, &end_ptr, 10 );
  if (end_ptr==str || *end_ptr!='\0' || long_val>INT_MAX || long_val<INT_MIN) {
    return 0;
  }
  return 1;
}

int send_signal_to_id( char* id, int do_stop ) {
  int child_pid;
  int ret_val = 0;
  if (!(child_pid=fork())) {
    if (quiet) {
      /* suppress warnings and messages from mupip */
      int fd = open("/dev/null", O_RDWR);
      if (-1 != fd) {
        dup2(fd, STDIN_FILENO);
        dup2(fd, STDOUT_FILENO);
        dup2(fd, STDERR_FILENO);
        close( fd );
      }
    }

    if (-1 == execl(mupip, mupip, do_stop ? "stop" : "intrpt", id, (char*)0)) {
      ERR_PRINT( "Error invoking mupip command.\n" );
      exit(1);
    }
  } else {
    int stat_loc;
    waitpid( child_pid, &stat_loc, 0 );
    if (WIFEXITED(stat_loc) && WEXITSTATUS(stat_loc)) {
      ret_val = 1;
    }
  }
  return ret_val;
}

int is_mumps_process( char *id ) {
  int ret_val;
  char pid_exe[PATH_MAX+1];
  char target[PATH_MAX+1];
  int size;
  sprintf( pid_exe, "/proc/%s/exe", id );
  size = readlink(pid_exe, target, PATH_MAX);
  target[size] = '\0';
  ret_val = !strcmp(target, mumps);
  if (!ret_val) {
    ERR_PRINT( "Processes %s is not a valid mumps process.\n", id );
  }
  return ret_val;
}

int send_signal_to_ids( char **ids, int do_stop ) {
  int i, num_failed;
  num_failed = 0;
  for (i=0; ids[i]; ++i) {
    if (is_mumps_process(ids[i])) {
      num_failed += send_signal_to_id( ids[i], do_stop );
    } else {
      ++num_failed;
    }
  }
  return num_failed;
}

int send_signal_to_all( int do_stop ) {
  struct dirent *dir_entry;
  FILE *fp;
  char *open_paren, *close_paren, *space;
  DIR *dir = opendir("/proc");
  int num_failed = 0;
  if (dir) {
    while ( dir_entry = readdir(dir) ) {
      if (!is_int(dir_entry->d_name)
          || !is_mumps_process(dir_entry->d_name)
          || !send_signal_to_id( dir_entry->d_name, do_stop )) {
        ++num_failed;
      }
    }
    closedir( dir );
  }
  return num_failed;
}

void find_paths() {
  char proc_exe[PATH_MAX];
  ssize_t size;
  int pid = getpid();
  struct stat status;
  sprintf(proc_exe, "/proc/%d/exe", pid);
  size = readlink(proc_exe, mupip, PATH_MAX+1);
  if (0>=size) {
    ERR_PRINT( "Could not determine where to find %s.\n", MUMPS_CMD );
    exit(1);
  }
  mupip[size]='\0';
  strcpy(mumps, mupip);
  strcpy(strrchr(mupip, '/')+1, MUPIP_CMD);
  strcpy(strrchr(mumps, '/')+1, MUMPS_CMD);
  if (-1 == stat(mupip, &status)) {
    ERR_PRINT( "Could not find '%s'.\n", mupip );
    exit(1);
  }
}

void print_usage(FILE *fp, const char* name) {
  fprintf( fp, "Usage: %s [OPTION] [pid]...\n", name );
  fprintf( fp, "Send mupip STOP or INTRPT to GT.M processes.\n" );
  fprintf( fp, "\nOptions\n" );
  fprintf( fp, "  -a, --all    send signal to all GT.M processes\n" );
  fprintf( fp, "  -s, --stop   send STOP instead of INTRPT\n" );
  fprintf( fp, "  -h, --help   print this help\n" );
  fprintf( fp, "  -q, --quiet  quiet; do not display warnings\n" );
}

int main( int argc, char **argv ) {
  int do_stop = 0;
  int all = 0;
  char **pids = 0;
  int array_index = 0;
  int num_failed = 0;

  find_paths();
  while (1) {
    static struct option long_options[] = 
    {
      {"stop", no_argument, 0, 's'},
      {"all", no_argument, 0, 'a'},
      {"quiet", no_argument, 0, 'q'},
      {"help", no_argument, 0, 'h'},
      {0, 0, 0, 0}
    };
    int c = getopt_long( argc, argv, "saqh", long_options, 0 );
    if (c==-1)
      break;
      switch (c) {
      case 's':
        do_stop = 1;
        break;
      case 'a':
        all = 1;
        break;
      case 'q':
        quiet = 1;
        break;
      case 'h':
        print_usage(stdout, argv[0]);
        exit(0);
      case '?':
        print_usage(stderr, argv[0]);
        exit(1);
      default:
        abort();
    }
  }

  if (optind < argc && !all) {
    pids = malloc((argc-optind+1)*sizeof(char*));
    pids[argc-optind] = 0;
    while (optind < argc) {
      if (!is_int(argv[optind])) {
        ERR_PRINT( "Invalid process number '%s'.\n", argv[optind] );
        exit(1);
      }
      pids[array_index++] = argv[optind];
      ++optind;
    }
  }

  if (all) {
    num_failed = send_signal_to_all( do_stop );
  } else if (pids) {
    num_failed = send_signal_to_ids( pids, do_stop );
    free( pids );
  } else {
    ERR_PRINT( "No process IDs were given.\n" );
    num_failed = 255;
  }

  return num_failed!=0;
}

